#ifndef DBOUNDARYLAYER_H
#define DBOUNDARYLAYER_H

#include "DClosures.h"
#include "blast.h"
#include "wObject.h"

#include <algorithm>

namespace blast
{

/**
 * @brief Boundary layer region upper/lower side or wake.
 * @author Paul Dechamps
 */
class BLAST_API BoundaryLayer : public fwk::wSharedObject
{

private:
    unsigned int const nVar =
        5;              ///< Number of variables of the partial differential problem.
    double nCrit = 9.0; ///< Critical amplification factor.

public:
    Closures *closSolver; ///< Closure relations class.
    std::string name;     ///< Name of the region.

    // Mesh
    size_t nNodes;             ///< Number of points in the domain.
    size_t nElms;              ///< Number of cells in the domain.
    std::vector<double> x;     ///< x coordinate in the global frame of reference.
    std::vector<double> y;     ///< y coordinate in the global frame of reference.
    std::vector<double> z;     ///< z coordinate in the global frame of reference.
    std::vector<double> xoc;   ///< x/c coordinate in the global frame of reference.
    std::vector<double> loc;   ///< xi coordinate in the local frame of reference.
    std::vector<int> rows;     ///< Reference numbers of the nodes.
    std::vector<int> no;       ///< Reference numbers of the nodes.
    std::vector<double> dx;    ///< Cell size.
    std::vector<double> alpha; ///< Angle of the cell wrt the x axis of the global
                               ///< frame of reference.
    double chord;              ///< Chord of the section.
    double xMin;               ///< Minimum x coordinate of the mesh (for local coordinates
                               ///< computation).

    // Boundary layer variables.
    std::vector<double> u; ///< Solution vector (theta, H, N, ue, Ct).
    std::vector<double>
        Ret;                ///< Reynolds number based on the momentum thickness (theta).
    std::vector<double> cd; ///< Local dissipation coefficient.
    std::vector<double> cf; ///< Local friction coefficient.
    std::vector<double>
        Hstar; ///< Kinetic energy shape parameter (thetaStar/theta).
    std::vector<double>
        Hstar2;             ///< Density shape parameter (deltaStarStar/theta).
    std::vector<double> Hk; ///< Kinematic shape parameter (int(1-u/ue d_eta)).
    std::vector<double>
        ctEq;                  ///< Equilibrium shear stress coefficient (turbulent BL).
    std::vector<double> us;    ///< Equivalent normalized wall slip velocity.
    std::vector<double> delta; ///< Boundary layer thickness.
    std::vector<double>
        deltaStar; ///< Displacement thickness (int(1-rho*u/rhoe*ue d_eta)).

    std::vector<double> Me;           ///< Mach number.
    std::vector<double> vt;           ///< Velocity.
    std::vector<double> rhoe;         ///< Density.
    std::vector<double> xxExt;        ///< xi coordinate in the local frame of reference,
                                      ///< at the edge of the boundary layer.
    std::vector<double> deltaStarExt; ///< Displacement thickness.
    std::vector<double> vtExt;        ///< Previous velocity.

    std::vector<double> blowingVelocity; ///< Blowing velocity.

    // Transition related variables.
    std::vector<int> regime; ///< Laminar (0) or turbulent (1) regime.
    size_t transitionNode;   ///< Node id of the transition location.
    double xtrF;             ///< Forced transition location in the local frame of reference
                             ///< (x/c).
    double xtr;              ///< Transition location in the local frame of reference (x/c).

    BoundaryLayer(double _xtrF = -1.0, std::string _name = "None");
    ~BoundaryLayer();

    void reset();

    // Getters
    std::string getName() const { return name; };
    size_t getnNodes() const { return nNodes; };
    size_t getnElms() const { return nElms; };

    std::vector<double> getDeltaStar() const { return deltaStar; };
    std::vector<double> getUe() const
    {
        std::vector<double> ue(nNodes, 0.);
        for (size_t i = 0; i < nNodes; ++i)
            ue[i] = u[i * nVar + 3];
        return ue;
    };
    std::vector<double> getBlowing() const { return blowingVelocity; };
    double getMaxMach() const { return *std::max_element(Me.begin(), Me.end()); };

    // Setters
    void setMesh(std::vector<double> const _x, std::vector<double> const y,
                 std::vector<double> const z, double const _chord,
                 double const _xMin,
                 std::vector<int> const _rows = std::vector<int>(0),
                 std::vector<int> const _no = std::vector<int>(0));
    void computeLocalCoordinates();
    void setMe(std::vector<double> const _Me)
    {
        if (_Me.size() != nNodes)
            throw std::runtime_error("Mach number vector is not consistent.");
        Me = _Me;
    };
    void setVt(std::vector<double> const _vt)
    {
        if (_vt.size() != nNodes)
            throw std::runtime_error("Velocity vector is not consistent.");
        vt = _vt;
    };
    void setRhoe(std::vector<double> const _rhoe)
    {
        if (_rhoe.size() != nNodes)
            throw std::runtime_error("Density vector is not consistent.");
        rhoe = _rhoe;
    };
    void setxxExt(std::vector<double> const _xxExt)
    {
        if (_xxExt.size() != nNodes)
            throw std::runtime_error("External mesh vector is not consistent.");
        xxExt = _xxExt;
    };
    void setDeltaStarExt(std::vector<double> const _deltaStarExt)
    {
        if (_deltaStarExt.size() != nNodes)
            throw std::runtime_error(
                "Displacement thickness vector is not consistent.");
        deltaStarExt = _deltaStarExt;
    };
    void setVtExt(std::vector<double> const _vtExt)
    {
        if (_vtExt.size() != nNodes)
            throw std::runtime_error("Velocity vector is not consistent.");
        vtExt = _vtExt;
    };
    void setxtrF(double const _xtrF) { xtrF = _xtrF; };

    // Boundary conditions.
    void setStagnationBC(double Re);
    void setWakeBC(double Re, std::vector<double> upperConditions,
                   std::vector<double> lowerConditions);

    // Getters.
    size_t getnVar() const { return nVar; };
    double getnCrit() const { return nCrit; };
    void setnCrit(double const _nCrit) { nCrit = _nCrit; };

    // Others
    void printSolution(size_t iPoint) const;
    void computeBlowingVelocity();
    std::vector<std::vector<double>> evalGradwrtRho();
    std::vector<std::vector<double>> evalGradwrtVt();
    std::vector<std::vector<double>> evalGradwrtDeltaStar();
    std::vector<std::vector<double>> evalGradwrtX();
    std::vector<std::vector<double>> evalGradwrtY();
};
} // namespace blast
#endif // DBOUNDARYLAYER_H
