/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "DCoupledAdjoint.h"
#include "DDriver.h"
#include "wAdjoint.h"
#include "wBlowing.h"
#include "wBlowingResidual.h"
#include "wFluid.h"
#include "wMshData.h"
#include "wNewton.h"
#include "wNode.h"
#include "wProblem.h"
#include "wSolver.h"

#include "wCache.h"
#include "wGauss.h"
#include "wGmres.h"
#include "wLinearSolver.h"

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <Eigen/SparseCholesky>
#include <Eigen/SparseLU>
#include <algorithm>
#include <chrono>
#include <deque>
#include <fstream>
#include <iomanip>

using namespace blast;

CoupledAdjoint::CoupledAdjoint(std::shared_ptr<dart::Adjoint> _iadjoint,
                               std::shared_ptr<blast::Driver> vSolver)
    : iadjoint(_iadjoint), vsol(vSolver)
{
    isol = iadjoint->sol;
    // Linear solvers (if GMRES, use the same, but with a thighter tolerance)
    if (isol->linsol->type() == tbox::LSolType::GMRES_ILUT)
    {
        std::shared_ptr<tbox::Gmres> gmres = std::make_shared<tbox::Gmres>(
            *dynamic_cast<tbox::Gmres *>(isol->linsol.get()));
        gmres->setTolerance(1e-8);
        alinsol = gmres;
    }
    else
        alinsol = isol->linsol;

    // Initalize all derivatives
    this->reset();
}

void CoupledAdjoint::reset()
{
    size_t nDim = isol->pbl->nDim;             // Number of dimensions of the problem
    size_t nNd = isol->pbl->msh->nodes.size(); // Number of domain nodes
    size_t nNs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nNs += bBC->nodes.size(); // Number of surface nodes
    nNs += 1;                     // Duplicate stagnation point
    size_t nEs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nEs += bBC->uE.size(); // Number of blowing elements

    dRphi_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nNd);
    dRM_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNd);
    dRrho_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNd);
    dRv_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nNd);
    dRdStar_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNd);
    dRblw_phi = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nNd);

    dRphi_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nNs);
    dRM_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRrho_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRv_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nNs);
    dRdStar_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRblw_M = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nNs);

    dRphi_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nNs);
    dRM_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRrho_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRv_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nNs);
    dRdStar_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRblw_rho = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nNs);

    dRphi_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nDim * nNs);
    dRM_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNs);
    dRrho_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNs);
    dRv_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nDim * nNs);
    dRdStar_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNs);
    dRblw_v = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nDim * nNs);

    dRphi_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nNs);
    dRM_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRrho_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRv_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nNs);
    dRdStar_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nNs);
    dRblw_dStar = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nNs);

    dRphi_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nEs);
    dRM_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nEs);
    dRrho_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nEs);
    dRv_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nEs);
    dRdStar_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nEs);
    dRblw_blw = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nEs);

    dRphi_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNd, nDim * nNd);
    dRM_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNd);
    dRrho_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNd);
    dRv_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNs, nDim * nNd);
    dRdStar_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nNs, nDim * nNd);
    dRblw_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nEs, nDim * nNd);

    // Gradient wrt objective function
    dRphi_AoA = Eigen::VectorXd::Zero(nNd);

    // Objective functions
    // Cl = f(phi) & Cd = Cdp + Cdf = f(phi, M, v)
    dCl_phi = Eigen::VectorXd::Zero(nNd);
    dCd_phi = Eigen::VectorXd::Zero(nNd);
    dCd_M = Eigen::VectorXd::Zero(nNs);
    dCd_v = Eigen::VectorXd::Zero(nDim * nNs);
    dCd_dStar = Eigen::VectorXd::Zero(nNs);

    // Angle of attack
    dCl_AoA = 0.0;
    dCd_AoA = 0.0;
    tdCl_AoA = 0.0;
    tdCd_AoA = 0.0;

    // Mesh
    dRx_x = Eigen::SparseMatrix<double, Eigen::RowMajor>(nDim * nNd, nDim * nNd);
    dCl_x = Eigen::VectorXd::Zero(nDim * nNd);
    dCdp_x = Eigen::VectorXd::Zero(nDim * nNd);
    dCdf_x = Eigen::VectorXd::Zero(nDim * nNd);

    tdCl_x.resize(nNd, Eigen::Vector3d::Zero());
    tdCd_x.resize(nNd, Eigen::Vector3d::Zero());
}

void CoupledAdjoint::buildAdjointMatrix(
    std::vector<std::vector<Eigen::SparseMatrix<double, Eigen::RowMajor> *>>
        &matrices,
    Eigen::SparseMatrix<double, Eigen::RowMajor> &matrixAdjoint)
{
    // Create a list of triplets for the larger matrix
    std::deque<Eigen::Triplet<double>> triplets;

    // Sanity check
    for (size_t irow = 0; irow < matrices.size(); ++irow)
    {
        if (irow > 0)
            if (matrices[irow].size() != matrices[irow - 1].size())
            {
                std::cout << "WRONG ROW SIZE " << std::endl;
                throw;
            }
        for (size_t icol = 0; icol < matrices[irow].size(); ++icol)
            if (icol > 0)
                if (matrices[irow][icol]->rows() != matrices[irow][icol - 1]->rows())
                {
                    std::cout << "WRONG COL SIZE " << std::endl;
                    throw;
                }
    }

    // Iterate over the rows and columns of the vector
    int rowOffset = 0;
    for (const auto &row : matrices)
    {
        int colOffset = 0;
        for (const auto &matrix : row)
        {
            // Add the triplets of the matrix to the list of triplets for the larger
            // matrix
            for (int k = 0; k < matrix->outerSize(); ++k)
            {
                for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(
                         *matrix, k);
                     it; ++it)
                {
                    triplets.push_back(Eigen::Triplet<double>(
                        it.row() + rowOffset, it.col() + colOffset, it.value()));
                }
            }
            colOffset += matrix->cols();
        }
        rowOffset += row[0]->rows();
    }
    // Create a new matrix from the deque of triplets
    matrixAdjoint.setFromTriplets(triplets.begin(), triplets.end());
    matrixAdjoint.prune(0.);
    matrixAdjoint.makeCompressed();
}

void CoupledAdjoint::run()
{
    tms["0-Total"].start();
    tms["1-Derivatives"].start();
    tms2["1-adj_inviscid"].start();
    gradientswrtInviscidFlow();
    tms2["1-adj_inviscid"].stop();
    tms2["2-adj_mach"].start();
    gradientswrtMachNumber();
    tms2["2-adj_mach"].stop();
    tms2["3-adj_density"].start();
    gradientswrtDensity();
    tms2["3-adj_density"].stop();
    tms2["4-adj_velocity"].start();
    gradientswrtVelocity();
    tms2["4-adj_velocity"].stop();
    tms2["5-adj_dStar"].start();
    gradientswrtDeltaStar();
    tms2["5-adj_dStar"].stop();
    tms2["6-adj_blw"].start();
    gradientswrtBlowingVelocity();
    tms2["6-adj_blw"].stop();
    tms2["7-adj_aoa"].start();
    gradientwrtAoA();
    tms2["7-adj_aoa"].stop();
    tms2["8-adj_mesh"].start();
    gradientwrtMesh();
    tms2["8-adj_mesh"].stop();
    tms["1-Derivatives"].stop();

    transposeMatrices();

    tms2["9-build"].start();
    std::vector<std::vector<Eigen::SparseMatrix<double, Eigen::RowMajor> *>>
        matrices = {{&dRphi_phi, &dRM_phi, &dRrho_phi, &dRv_phi, &dRdStar_phi,
                     &dRblw_phi},
                    {&dRphi_M, &dRM_M, &dRrho_M, &dRv_M, &dRdStar_M, &dRblw_M},
                    {&dRphi_rho, &dRM_rho, &dRrho_rho, &dRv_rho, &dRdStar_rho,
                     &dRblw_rho},
                    {&dRphi_v, &dRM_v, &dRrho_v, &dRv_v, &dRdStar_v, &dRblw_v},
                    {&dRphi_dStar, &dRM_dStar, &dRrho_dStar, &dRv_dStar,
                     &dRdStar_dStar, &dRblw_dStar},
                    {&dRphi_blw, &dRM_blw, &dRrho_blw, &dRv_blw, &dRdStar_blw,
                     &dRblw_blw}};

    int rows = 0;
    int cols = 0;
    for (const auto &row : matrices)
        rows +=
            row[0]->rows(); // All matrices in a row have the same number of rows
    for (const auto &mat1 : matrices[0])
        cols += mat1->cols(); // All matrices in a column have the same number of
                              // columns
    Eigen::SparseMatrix<double, Eigen::RowMajor> A_adjoint(rows, cols);
    buildAdjointMatrix(matrices, A_adjoint);
    tms2["9-build"].stop();

    size_t phiIdx = 0;
    size_t machIdx = dRphi_phi.cols();
    size_t rhoIdx = machIdx + dRM_M.cols();
    size_t vIdx = rhoIdx + dRrho_rho.cols();
    size_t dStarIdx = vIdx + dRv_v.cols();
    size_t blwIdx = dStarIdx + dRdStar_dStar.cols();

    //**************************************************************************//
    //                      Gradients wrt angle of attack                       //
    //**************************************************************************//
    // CL
    tms["2-Solve"].start();
    tms2["10-solveCxAoA"].start();
    std::vector<double> rhsCl(A_adjoint.cols(), 0.0);
    for (auto i = phiIdx; i < phiIdx + dCl_phi.size(); ++i)
        rhsCl[i] = dCl_phi[i]; // Only dCl_dphi is non-zero
    Eigen::VectorXd rhsCl_ =
        Eigen::Map<const Eigen::VectorXd>(rhsCl.data(), rhsCl.size());

    Eigen::VectorXd lambdaCl(A_adjoint.cols()); // Solution vector for lambdasCl
    Eigen::Map<Eigen::VectorXd> lambdaCl_(lambdaCl.data(), lambdaCl.size());

    alinsol->compute(A_adjoint,
                     Eigen::Map<Eigen::VectorXd>(rhsCl_.data(), rhsCl_.size()),
                     lambdaCl_);

    // Total gradient
    tdCl_AoA = dCl_AoA -
               dRphi_AoA.transpose() * lambdaCl.segment(phiIdx, dRphi_phi.cols());

    // CD
    std::vector<double> rhsCd(A_adjoint.cols(), 0.0);
    int jj = 0;
    for (auto i = phiIdx; i < phiIdx + dCd_phi.size(); ++i)
    {
        rhsCd[i] = dCd_phi[jj];
        jj++;
    }
    jj = 0;
    for (auto i = machIdx; i < machIdx + dCd_M.size(); ++i)
    {
        rhsCd[i] = dCd_M[jj];
        jj++;
    }
    jj = 0;
    for (auto i = vIdx; i < vIdx + dCd_v.size(); ++i)
    {
        rhsCd[i] = dCd_v[jj];
        jj++;
    }
    jj = 0;
    for (auto i = dStarIdx; i < dStarIdx + dCd_dStar.size(); ++i)
    {
        rhsCd[i] = dCd_dStar[jj];
        jj++;
    }
    Eigen::VectorXd rhsCd_ =
        Eigen::Map<const Eigen::VectorXd>(rhsCd.data(), rhsCd.size());

    Eigen::VectorXd lambdaCd(A_adjoint.cols()); // Solution vector for lambdasCd
    Eigen::Map<Eigen::VectorXd> lambdaCd_(lambdaCd.data(), lambdaCd.size());

    alinsol->compute(A_adjoint,
                     Eigen::Map<Eigen::VectorXd>(rhsCd_.data(), rhsCd_.size()),
                     lambdaCd_);

    // Total gradient
    tdCd_AoA = dCd_AoA -
               dRphi_AoA.transpose() *
                   lambdaCd.segment(phiIdx, dRphi_phi.cols()); // Total gradient
    tms2["10-solveCxAoA"].stop();

    //**************************************************************************//
    //                     Gradients wrt mesh coordinates                       //
    //--------------------------------------------------------------------------//
    // Ck (cL, cd)                                                              //
    // Solution lambdaCl_x of (from augmented Lagrangian, eqn d/dx )            //
    // "dCk_x + dRphi_x * lambdaCk_phi + dRdStar_x * lambdaCk_x                 //
    // + dRblw_x * lambdaCk_blw + dRx_x * lambdaCk_x = 0"                       //
    //**************************************************************************//

    tms2["12-solveCxMesh"].start();
    Eigen::VectorXd rhsCl_x = dCl_x;
    rhsCl_x -=
        dRphi_x.transpose() *
        lambdaCl.segment(phiIdx, dRphi_phi.cols()); // Potential contribution
    rhsCl_x -=
        dRM_x.transpose() *
        lambdaCl.segment(machIdx, dRM_M.cols()); // Mach number contribution
    rhsCl_x -= dRrho_x.transpose() *
               lambdaCl.segment(rhoIdx, dRrho_rho.cols()); // Density contribution
    rhsCl_x -= dRv_x.transpose() *
               lambdaCl.segment(vIdx, dRv_v.cols()); // Velocity contribution
    rhsCl_x -= dRdStar_x.transpose() *
               lambdaCl.segment(
                   dStarIdx,
                   dRdStar_dStar.cols()); // Displacement thickness contribution
    rhsCl_x -= dRblw_x.transpose() *
               lambdaCl.segment(blwIdx, dRblw_blw.cols()); // Blowing contribution

    Eigen::VectorXd lambdaCl_x =
        Eigen::VectorXd::Zero(isol->pbl->nDim * isol->pbl->msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> lambdaCl_x_(lambdaCl_x.data(), lambdaCl_x.size());
    alinsol->compute(dRx_x.transpose(),
                     Eigen::Map<Eigen::VectorXd>(rhsCl_x.data(), rhsCl_x.size()),
                     lambdaCl_x_);

    Eigen::VectorXd rhsCd_x = dCdp_x; // Pressure drag coefficient contribution
    rhsCd_x += dCdf_x;                // Friction drag coefficient contribution
    rhsCd_x -=
        dRphi_x.transpose() *
        lambdaCd.segment(phiIdx, dRphi_phi.cols()); // Potential contribution
    rhsCd_x -=
        dRM_x.transpose() *
        lambdaCd.segment(machIdx, dRM_M.cols()); // Mach number contribution
    rhsCd_x -= dRrho_x.transpose() *
               lambdaCd.segment(rhoIdx, dRrho_rho.cols()); // Density contribution
    rhsCd_x -= dRv_x.transpose() *
               lambdaCd.segment(vIdx, dRv_v.cols()); // Velocity contribution
    rhsCd_x -= dRdStar_x.transpose() *
               lambdaCd.segment(
                   dStarIdx,
                   dRdStar_dStar.cols()); // Displacement thickness contribution
    rhsCd_x -= dRblw_x.transpose() *
               lambdaCd.segment(blwIdx, dRblw_blw.cols()); // Blowing contribution

    Eigen::VectorXd lambdaCd_x =
        Eigen::VectorXd::Zero(isol->pbl->nDim * isol->pbl->msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> lambdaCd_x_(lambdaCd_x.data(), lambdaCd_x.size());
    alinsol->compute(dRx_x.transpose(),
                     Eigen::Map<Eigen::VectorXd>(rhsCd_x.data(), rhsCd_x.size()),
                     lambdaCd_x_);

    for (auto n : isol->pbl->msh->nodes)
    {
        for (int m = 0; m < isol->pbl->nDim; m++)
        {
            tdCl_x[n->row](m) = lambdaCl_x[isol->pbl->nDim * (n->row) + m];
            tdCd_x[n->row](m) = lambdaCd_x[isol->pbl->nDim * (n->row) + m];
        }
    }
    tms2["12-solveCxMesh"].stop();
    tms["2-Solve"].stop();

    // Print output
    std::cout << "-------------- Adjoint solution --------------" << std::endl;
    // aoa gradients
    std::cout << std::setw(15) << std::left << "Adjoint AoA" << std::setw(15)
              << std::right << "Res[Cl_AoA]" << std::setw(15) << std::right
              << "Res[Cd_AoA]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << "" << std::setw(15) << std::right
              << log10((A_adjoint * Eigen::Map<Eigen::VectorXd>(lambdaCl.data(),
                                                                lambdaCl.size()) -
                        Eigen::Map<Eigen::VectorXd>(rhsCl.data(), rhsCl.size()))
                           .norm())
              << std::setw(15) << std::right
              << log10((A_adjoint * Eigen::Map<Eigen::VectorXd>(lambdaCd.data(),
                                                                lambdaCd.size()) -
                        Eigen::Map<Eigen::VectorXd>(rhsCd.data(), rhsCd.size()))
                           .norm())
              << std::endl;
    std::cout << std::setw(15) << std::left << "AoA gradients" << std::setw(15)
              << std::right << "dCl_dAoA" << std::setw(15) << std::right
              << "dCd_dAoA" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << "" << std::setw(15) << std::right
              << tdCl_AoA << std::setw(15) << std::right << tdCd_AoA << std::endl;
    // mesh gradients
    std::cout << std::setw(15) << std::left << "Adjoint mesh" << std::setw(15)
              << std::right << "Res[Cl_x]" << std::setw(15) << std::right
              << "Res[Cd_x]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout
        << std::setw(15) << std::left << "" << std::setw(15) << std::right
        << log10((dRx_x.transpose() * Eigen::Map<Eigen::VectorXd>(
                                          lambdaCl_x.data(), lambdaCl_x.size()) -
                  Eigen::Map<Eigen::VectorXd>(rhsCl_x.data(), rhsCl_x.size()))
                     .norm())
        << std::setw(15) << std::right
        << log10((dRx_x.transpose() * Eigen::Map<Eigen::VectorXd>(
                                          lambdaCd_x.data(), lambdaCd_x.size()) -
                  Eigen::Map<Eigen::VectorXd>(rhsCd_x.data(), rhsCd_x.size()))
                     .norm())
        << std::endl;
    std::cout << std::setw(15) << std::left << "Mesh gradients" << std::setw(15)
              << std::right << "Norm[dCl_dX]" << std::setw(15) << std::right
              << "Norm[dCd_dX]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << "" << std::setw(15) << std::right
              << Eigen::Map<Eigen::VectorXd>(lambdaCl_x.data(), lambdaCl_x.size())
                     .norm()
              << std::setw(15) << std::right
              << Eigen::Map<Eigen::VectorXd>(lambdaCd_x.data(), lambdaCd_x.size())
                     .norm()
              << std::endl;
    tms["0-Total"].stop();
    std::cout << "--------------- Adjoint timers ---------------" << std::endl;
    std::cout << tms << "----------------------------------------------"
              << std::endl;
    if (vsol->verbose > 2)
        std::cout << tms2 << "----------------------------------------------"
                  << std::endl;
}

/**
 * @brief Compute all the gradients wrt the inviscid flow (phi) in the domain
 * Non-zero are: dRphi_phi, dRM_phi, dRrho_phi, dRv_phi
 * Zeros are: dRdStar_phi, dRblw_phi, dRx_phi
 */
void CoupledAdjoint::gradientswrtInviscidFlow()
{
    //**************************************************************************//
    //                                 dRphi_phi                                //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    dRphi_phi = iadjoint->getRu_U();

    //**************************************************************************//
    //                      dRM_phi, dRrho_phi, dRv_phi                         //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    auto pbl = isol->pbl;

    std::deque<Eigen::Triplet<double>> T_dM;
    std::deque<Eigen::Triplet<double>> T_drho;
    std::deque<Eigen::Triplet<double>> T_dv;

    size_t jj = 0;
    int iReg = 0;
    for (auto sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error(
                "gradientswrtInviscidFlow: Unknown section name");
        for (size_t iNode = 0; iNode < sec->nNodes; ++iNode)
        {
            tbox::Node *srfNode = pbl->bBCs[iReg]->nodes[sec->rows[iNode]];
            double nAdjElms = static_cast<double>(pbl->fluid->neMap[srfNode].size());
            for (auto e : pbl->fluid->neMap[srfNode])
            {
                // dv/dphi = dv/dgradPhi * dgradPhi/dphi = dgradPhi/dphi = d(gradN_k
                // phi_k) / dphi_j dM/dphi = dM/dgradPhi * dgradPhi/dphi = dM/dgradPhi *
                // dv/dphi drho/dphi = drho/dgradPhi * dgradPhi/dphi = drho/dgradPhi *
                // dv/dphi
                Eigen::MatrixXd gradV = e->getJinv(0) * e->getVCache().getDsf(0);
                Eigen::VectorXd gradM =
                    isol->pbl->fluid->mach->evalGrad(*e, isol->phi, 0) *
                    e->computeGradient(isol->phi, 0).transpose() * gradV;
                Eigen::VectorXd gradrho =
                    isol->pbl->fluid->rho->evalGrad(*e, isol->phi, 0) *
                    e->computeGradient(isol->phi, 0).transpose() * gradV;
                for (size_t k = 0; k < e->nodes.size(); ++k)
                {
                    T_dM.push_back(Eigen::Triplet<double>(jj, e->nodes[k]->row,
                                                          -gradM(k) / nAdjElms));
                    T_drho.push_back(Eigen::Triplet<double>(jj, e->nodes[k]->row,
                                                            -gradrho(k) / nAdjElms));
                    for (int idim = 0; idim < pbl->nDim; ++idim)
                        T_dv.push_back(Eigen::Triplet<double>(
                            jj * pbl->nDim + idim, e->nodes[k]->row,
                            -gradV(k * pbl->nDim + idim) / nAdjElms));
                }
            }
            ++jj;
        }
    }
    dRM_phi.setFromTriplets(T_dM.begin(), T_dM.end());
    dRrho_phi.setFromTriplets(T_drho.begin(), T_drho.end());
    dRv_phi.setFromTriplets(T_dv.begin(), T_dv.end());
    dRM_phi.prune(0.);
    dRrho_phi.prune(0.);
    dRv_phi.prune(0.);

    //**************************************************************************//
    //                             dCl_phi, dCd_phi                             //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    std::vector<std::vector<double>> dCx_phi =
        iadjoint->computeGradientCoefficientsFlow();
    dCl_phi = Eigen::VectorXd::Map(dCx_phi[0].data(), dCx_phi[0].size());
    dCd_phi = Eigen::VectorXd::Map(dCx_phi[1].data(), dCx_phi[1].size());
}

/**
 * @brief Compute all the gradients wrt the Mach number on the surface
 * Non-zero are: dRM_M, dRdStar_M
 * Zeros are: dRphi_M, dRrho_M, dRv_M, dRblw_M, dRx_M
 */
void CoupledAdjoint::gradientswrtMachNumber()
{
    //**************************************************************************//
    //                                   dRM_M                                  //
    //**************************************************************************//
    dRM_M.setIdentity();

    //**************************************************************************//
    //                                 dRdStar_M                                //
    //--------------------------------------------------------------------------//
    //                         Central finite-difference                        //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T;
    double delta = 1e-6;
    size_t i = 0;
    double saveM = 0.;
    std::vector<Eigen::VectorXd> ddStar(2,
                                        Eigen::VectorXd::Zero(dRdStar_M.cols()));
    std::vector<double> dCdf(2, 0.);
    for (auto sec : vsol->sections[0])
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            saveM = sec->Me[j];

            sec->Me[j] = saveM + delta;
            ddStar[0] = this->runViscous();
            dCdf[0] = vsol->Cdf;
            sec->Me[j] = saveM - delta;
            ddStar[1] = this->runViscous();
            dCdf[1] = vsol->Cdf;

            sec->Me[j] = saveM;

            for (int k = 0; k < ddStar[0].size(); ++k)
                T.push_back(Eigen::Triplet<double>(
                    k, i, -(ddStar[0][k] - ddStar[1][k]) / (2. * delta)));
            // Collect Cd gradients
            dCd_M(i) = (dCdf[0] - dCdf[1]) / (2. * delta);
            ++i;
        }
    dRdStar_M.setFromTriplets(T.begin(), T.end());
    dRdStar_M.prune(0.);
}

/**
 * @brief Compute all the gradients wrt the density on the surface
 * Non-zero are: dRrho_rho, dRblw_rho
 * Zeros are: dRphi_rho, dRM_rho, dRv_rho, dRdStar_rho, dRx_rho
 */
void CoupledAdjoint::gradientswrtDensity()
{
    size_t nNs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nNs += bBC->nodes.size(); // Number of surface nodes
    nNs += 1;                     // Duplicate stagnation point
    size_t nEs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nEs += bBC->uE.size(); // Number of blowing elements

    //**************************************************************************//
    //                                 dRrho_rho                                //
    //**************************************************************************//
    dRrho_rho.setIdentity();

    //**************************************************************************//
    //                                 dRblw_rho                                //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T_blw;
    int offSetElms = 0;
    int offSetNodes = 0;
    for (const auto &sec : vsol->sections[0])
    {
        std::vector<std::vector<double>> grad = sec->evalGradwrtRho();
        for (size_t i = 0; i < grad.size(); ++i)
            for (size_t j = 0; j < grad[i].size(); ++j)
                T_blw.push_back(Eigen::Triplet<double>(i + offSetElms, j + offSetNodes,
                                                       -grad[i][j]));
        offSetElms += sec->nElms;
        offSetNodes += sec->nNodes;
    }
    dRblw_rho.setFromTriplets(T_blw.begin(), T_blw.end());
    dRblw_rho.prune(0.);
}

/**
 * @brief Compute all the gradients wrt the velocity on the surface
 * Non-zeros are: dRv_v, dRdStar_v, dRblw_v
 * Zeros are: dRphi_v, dRM_v, dRrho_v, dRx_v
 */
void CoupledAdjoint::gradientswrtVelocity()
{
    size_t nDim = isol->pbl->nDim; // Number of dimensions of the problem
    size_t nNs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nNs += bBC->nodes.size(); // Number of surface nodes
    nNs += 1;                     // Duplicate stagnation point
    size_t nEs = 0;
    for (auto bBC : isol->pbl->bBCs)
        nEs += bBC->uE.size(); // Number of blowing elements

    //**************************************************************************//
    //                                   dRv_v                                  //
    //**************************************************************************//
    dRv_v.setIdentity();

    //**** Get velocity****/
    Eigen::MatrixXd v = Eigen::MatrixXd::Zero(nNs, nDim);
    size_t i = 0;
    int iReg = -1;
    for (const auto &sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error("Unknown section name");
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            for (size_t idim = 0; idim < nDim; ++idim)
                v(i, idim) =
                    isol->U[isol->pbl->bBCs[iReg]->nodes[sec->rows[j]]->row](idim);
            ++i;
        }
    }

    //**************************************************************************//
    //                                 dvt_v                                    //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    Eigen::SparseMatrix<double, Eigen::RowMajor> dVt_v(nNs, nNs * nDim);
    std::deque<Eigen::Triplet<double>> T_vt;

    i = 0;
    for (const auto &sec : vsol->sections[0])
    {
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            for (size_t idim = 0; idim < nDim; ++idim)
            {
                T_vt.push_back(Eigen::Triplet<double>(
                    i, i * nDim + idim,
                    1 / std::sqrt(v(i, 0) * v(i, 0) + v(i, 1) * v(i, 1)) * v(i, idim)));
            }
            ++i;
        }
    }
    dVt_v.setFromTriplets(T_vt.begin(), T_vt.end());
    dVt_v.prune(0.);

    //**************************************************************************//
    //                                 dRdStar_vt                               //
    //--------------------------------------------------------------------------//
    //                          Central finite-difference                       //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T;
    double delta = 1e-6;
    i = 0;
    double saveM = 0.;
    std::vector<Eigen::VectorXd> dvt(2, Eigen::VectorXd::Zero(dRdStar_M.cols()));
    std::vector<double> dCdf(2, 0.);
    Eigen::VectorXd dCd_vt = Eigen::VectorXd::Zero(dRM_M.cols());
    for (auto sec : vsol->sections[0])
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            saveM = sec->vt[j];

            sec->vt[j] = saveM + delta;
            dvt[0] = this->runViscous();
            dCdf[0] = vsol->Cdf;
            sec->vt[j] = saveM - delta;
            dvt[1] = this->runViscous();
            dCdf[1] = vsol->Cdf;

            sec->vt[j] = saveM;

            for (int k = 0; k < dvt[0].size(); ++k)
                T.push_back(Eigen::Triplet<double>(
                    k, i, -(dvt[0][k] - dvt[1][k]) / (2. * delta)));
            dCd_vt(i) = (dCdf[0] - dCdf[1]) / (2. * delta);
            ++i;
        }
    Eigen::SparseMatrix<double, Eigen::RowMajor> dRdStar_vt(dRdStar_M.rows(),
                                                            dRdStar_M.cols());
    dRdStar_vt.setFromTriplets(T.begin(), T.end());
    dRdStar_vt.prune(0.);

    //**************************************************************************//
    //                                 dRdStar_v                                //
    //**************************************************************************//
    dRdStar_v = dRdStar_vt * dVt_v;
    dRdStar_v.prune(0.);

    //**************************************************************************//
    //                                   dCdf_v                                 //
    //**************************************************************************//
    dCd_v =
        dCd_vt.transpose() * dVt_v; // [1xnNs] x [nNs x nNs*nDim] = [1 x nNs*nDim]

    //**************************************************************************//
    //                                 dRblw_vt                                 //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T_blw;
    int offSetElms = 0;
    int offSetNodes = 0;
    for (const auto &sec : vsol->sections[0])
    {
        std::vector<std::vector<double>> grad = sec->evalGradwrtVt();
        for (size_t i = 0; i < grad.size(); ++i)
            for (size_t j = 0; j < grad[i].size(); ++j)
                T_blw.push_back(Eigen::Triplet<double>(i + offSetElms, j + offSetNodes,
                                                       -grad[i][j]));
        offSetElms += sec->nElms;
        offSetNodes += sec->nNodes;
    }

    Eigen::SparseMatrix<double, Eigen::RowMajor> dRblw_vt(nEs, nNs);
    dRblw_vt.setFromTriplets(T_blw.begin(), T_blw.end());
    dRblw_vt.prune(0.);

    //**************************************************************************//
    //                                   dRblw_v                                //
    //**************************************************************************//
    dRblw_v = dRblw_vt * dVt_v;
    dRblw_v.prune(0.);
}

/**
 * @brief Compute all the gradients wrt the displacement thickness on the
 * surface Non-zero are: dRdStar_dStar, dRblw_dStar Zeros are: dRphi_dStar,
 * dRM_dStar, dRrho_dStar, dRv_dStar, dRx_dStar
 */
void CoupledAdjoint::gradientswrtDeltaStar()
{
    //**************************************************************************//
    //                              dRdStar_dStar                               //
    //--------------------------------------------------------------------------//
    //                        Central finite-difference                         //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T;
    double delta = 1e-6;
    double saveDStar = 0.;
    std::vector<Eigen::VectorXd> ddstar(
        2, Eigen::VectorXd::Zero(dRdStar_dStar.cols()));
    std::vector<double> dCdf(2, 0.);
    size_t i = 0;
    for (auto sec : vsol->sections[0])
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            saveDStar = sec->deltaStarExt[j];

            sec->deltaStarExt[j] = saveDStar + delta;
            ddstar[0] = this->runViscous();
            dCdf[0] = vsol->Cdf;
            sec->deltaStarExt[j] = saveDStar - delta;
            ddstar[1] = this->runViscous();
            dCdf[1] = vsol->Cdf;

            sec->deltaStarExt[j] = saveDStar;

            for (int k = 0; k < ddstar[0].size(); ++k)
                T.push_back(Eigen::Triplet<double>(
                    k, i, (ddstar[0][k] - ddstar[1][k]) / (2. * delta)));
            dCd_dStar(i) = (dCdf[0] - dCdf[1]) / (2. * delta);
            ++i;
        }
    Eigen::SparseMatrix<double, Eigen::RowMajor> dRdStar_dStar_dum(
        dRdStar_dStar.rows(), dRdStar_dStar.cols());
    // RdStar = dStar - f(dStar) = 0
    // dRdStar_dStar = 1 - df_dStar (minus is 3 lines below and not in the
    // triplets)
    dRdStar_dStar_dum.setFromTriplets(T.begin(), T.end());
    dRdStar_dStar.setIdentity();
    dRdStar_dStar -= dRdStar_dStar_dum;
    dRdStar_dStar.prune(0.);

    //**************************************************************************//
    //                                 dRblw_dStar                              //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T_blw;
    int offSetElms = 0;
    int offSetNodes = 0;
    for (const auto &sec : vsol->sections[0])
    {
        std::vector<std::vector<double>> grad = sec->evalGradwrtDeltaStar();
        for (size_t i = 0; i < grad.size(); ++i)
            for (size_t j = 0; j < grad[i].size(); ++j)
                T_blw.push_back(Eigen::Triplet<double>(i + offSetElms, j + offSetNodes,
                                                       -grad[i][j]));
        offSetElms += sec->nElms;
        offSetNodes += sec->nNodes;
    }
    dRblw_dStar.setFromTriplets(T_blw.begin(), T_blw.end());
    dRblw_dStar.prune(0.);
}

/**
 * @brief Compute all the gradients wrt the blowing velocity
 * Non-zero are: dRphi_blw, dRblw_blw
 * Zeros are: dRM_blw, dRrho_blw, dRv_blw, dRdStar_blw, dRx_blw
 */
void CoupledAdjoint::gradientswrtBlowingVelocity()
{
    //**************************************************************************//
    //                                 dRphi_blw                                //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T;

    size_t jj = 0;
    int iReg = 0;
    for (auto sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error("Unknown section name");
        for (size_t iElm = 0; iElm < sec->nElms; ++iElm)
        {
            auto blowingElement = isol->pbl->bBCs[iReg]->uE[sec->no[iElm]].first;
            Eigen::VectorXd be =
                dart::BlowingResidual::buildGradientBlowing(*blowingElement);
            for (size_t ii = 0; ii < blowingElement->nodes.size(); ii++)
            {
                tbox::Node *nodi = blowingElement->nodes[ii];
                T.push_back(
                    Eigen::Triplet<double>(isol->getRow(nodi->row), jj, be(ii)));
            }
            ++jj;
        }
    }
    dRphi_blw.setFromTriplets(T.begin(), T.end());
    dRphi_blw.prune(0.);
    dRphi_blw.makeCompressed();

    /**** dRblw_blw ****/
    dRblw_blw.setIdentity();
}

/**
 * @brief Compute all the gradients wrt the angle of attack
 * Non-zero are: dRphi_AoA
 * Zeros are: dRM_AoA, dRrho_AoA, dRv_AoA, dRdStar_AoA, dRblw_AoA, dRx_AoA
 */
void CoupledAdjoint::gradientwrtAoA()
{
    //**************************************************************************//
    //                              dCl_Aoa dCd_Aoa                             //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    std::vector<double> dCx_AoA = iadjoint->computeGradientCoefficientsAoa();
    dCl_AoA = dCx_AoA[0];
    dCd_AoA = dCx_AoA[1];

    //**************************************************************************//
    //                                 dRphi_Aoa                                //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    dRphi_AoA = iadjoint->getRu_A();
}

/**
 * @brief Compute all the gradients wrt the mesh coordinates
 * Non-zero are: dRphi_x, dRM_x, dRrho_x, dRv_x, dRdStar_x, dRblw_x, dRx_x
 * Zeros are: -
 */
void CoupledAdjoint::gradientwrtMesh()
{

    int nDim = isol->pbl->nDim;

    //**************************************************************************//
    //                               dCl_x, dCd_x                               //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    std::vector<std::vector<double>> dCx_x =
        iadjoint->computeGradientCoefficientsMesh();
    dCl_x = Eigen::Map<Eigen::VectorXd>(dCx_x[0].data(), dCx_x[0].size());
    dCdp_x = Eigen::Map<Eigen::VectorXd>(dCx_x[1].data(), dCx_x[1].size());

    //**************************************************************************//
    //                                  dRphi_x                                 //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    dRphi_x = iadjoint->getRu_X();
    //**************************************************************************//
    //                                   dRx_x                                  //
    //--------------------------------------------------------------------------//
    //                                 Analytical                               //
    //**************************************************************************//
    dRx_x = iadjoint->getRx_X();

    //**************************************************************************//
    //                            dRdStar_x, dCdf_dx                            //
    //--------------------------------------------------------------------------//
    //                         Central finite-difference                        //
    //--------------------------------------------------------------------------//
    // We don't do dRdStar_x = dRdStar_loc * dloc_dx because Cdf = f(xi(x), x)  //
    // and we would have to compute dCdf/dx and dCdf/dloc. Same for deltaStar.  //
    // It is more computationally expansive to compute                          //
    // ddeltaStar/dx, ddeltaStar/dloc than ddeltaStar_dx, ddeltaStar_dy         //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T_dStar_x;
    double delta = 1e-8;
    int iReg = 0;
    std::vector<Eigen::VectorXd> ddStar(2,
                                        Eigen::VectorXd::Zero(dRdStar_M.cols()));
    std::vector<double> dCdf(2, 0.);
    for (auto sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error("Unknown section name");
        for (size_t j = 0; j < sec->nNodes; ++j)
        {
            int rowj = isol->pbl->bBCs[iReg]->nodes[sec->rows[j]]->row;
            // x
            double saveX = sec->x[j];

            sec->x[j] = saveX + delta;
            sec->computeLocalCoordinates();
            sec->xxExt = sec->loc;
            ddStar[0] = this->runViscous();
            dCdf[0] = vsol->Cdf;
            sec->x[j] = saveX - delta;
            sec->computeLocalCoordinates();
            sec->xxExt = sec->loc;
            ddStar[1] = this->runViscous();
            dCdf[1] = vsol->Cdf;

            sec->x[j] = saveX;
            sec->xxExt = sec->loc;
            sec->computeLocalCoordinates();

            for (int k = 0; k < ddStar[0].size(); ++k)
                T_dStar_x.push_back(Eigen::Triplet<double>(
                    k, rowj * isol->pbl->nDim + 0,
                    -(ddStar[0][k] - ddStar[1][k]) / (2. * delta)));
            dCdf_x(rowj * isol->pbl->nDim + 0) += (dCdf[0] - dCdf[1]) / (2. * delta);
            // y
            double saveY = sec->y[j];

            sec->y[j] = saveY + delta;
            sec->computeLocalCoordinates();
            sec->xxExt = sec->loc;
            ddStar[0] = this->runViscous();
            dCdf[0] = vsol->Cdf;
            sec->y[j] = saveY - delta;
            sec->computeLocalCoordinates();
            sec->xxExt = sec->loc;
            ddStar[1] = this->runViscous();
            dCdf[1] = vsol->Cdf;

            sec->y[j] = saveY;
            sec->xxExt = sec->loc;
            sec->computeLocalCoordinates();

            for (int k = 0; k < ddStar[0].size(); ++k)
                T_dStar_x.push_back(Eigen::Triplet<double>(
                    k, rowj * isol->pbl->nDim + 1,
                    -(ddStar[0][k] - ddStar[1][k]) / (2. * delta)));
            dCdf_x(rowj * isol->pbl->nDim + 1) += (dCdf[0] - dCdf[1]) / (2. * delta);
        }
    }
    dRdStar_x.setFromTriplets(T_dStar_x.begin(), T_dStar_x.end());
    dRdStar_x.prune(0.);

    //**************************************************************************//
    //                                dRblw_x                                   //
    //--------------------------------------------------------------------------//
    //                               Analytical                                 //
    //**************************************************************************//
    std::deque<Eigen::Triplet<double>> T_blw_x;
    int offSetElms = 0;
    for (const auto &sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error("Unknown section name");
        std::vector<std::vector<double>> gradX = sec->evalGradwrtX();
        std::vector<std::vector<double>> gradY = sec->evalGradwrtY();
        for (size_t i = 0; i < gradX.size(); ++i)
            for (size_t j = 0; j < gradX[i].size(); ++j)
            {
                int rowj = isol->pbl->bBCs[iReg]->nodes[sec->rows[j]]->row;
                T_blw_x.push_back(Eigen::Triplet<double>(
                    i + offSetElms, rowj * isol->pbl->nDim + 0, -gradX[i][j]));
                T_blw_x.push_back(Eigen::Triplet<double>(
                    i + offSetElms, rowj * isol->pbl->nDim + 1, -gradY[i][j]));
            }
        offSetElms += sec->nElms;
    }
    dRblw_x.setFromTriplets(T_blw_x.begin(), T_blw_x.end());
    dRblw_x.prune(0.);

    //**************************************************************************//
    //                         dRM_x, dRrho_x, dRv_x                            //
    //--------------------------------------------------------------------------//
    //                              Analytical                                  //
    //**************************************************************************//
    auto pbl = isol->pbl;

    std::deque<Eigen::Triplet<double>> T_dM_x;
    std::deque<Eigen::Triplet<double>> T_drho_x;
    std::deque<Eigen::Triplet<double>> T_dv_x;

    int jj = 0;
    iReg = 0;

    for (auto sec : vsol->sections[0])
    {
        if (sec->name == "wake")
            iReg = 1;
        else if (sec->name == "upper" || sec->name == "lower")
            iReg = 0;
        else
            throw std::runtime_error(
                "gradientswrtInviscidFlow: Unknown section name");
        for (size_t iNode = 0; iNode < sec->nNodes; ++iNode)
        {
            tbox::Node *srfNode = pbl->bBCs[iReg]->nodes[sec->rows[iNode]];
            double nAdjElms = static_cast<double>(pbl->fluid->neMap[srfNode].size());
            for (auto e : pbl->fluid->neMap[srfNode])
            {
                // dv/dx = dv/dgradPhi * dgradPhi/dx = dgradPhi/dx = (-J^(-1) *
                // dJ/dx_j)*grad_xPhi dM/dx = dM/dgradPhi * dgradPhi/dx = dM/dgradPhi *
                // gradPhi * dv/dx drho/dx = drho/dgradPhi * dgradPhi/dx = drho/dgradPhi
                // * gradPhi * dv/dx
                for (size_t inodOnElm = 0; inodOnElm < e->nodes.size(); inodOnElm++)
                    for (int idim = 0; idim < nDim; ++idim)
                    {
                        Eigen::VectorXd gradV = (-e->getJinv(0)) *
                                                e->getGradJ(0)[inodOnElm * nDim + idim] *
                                                e->computeGradient(isol->phi, 0);
                        double gradM = pbl->fluid->mach->evalGrad(*e, isol->phi, 0) *
                                       e->computeGradient(isol->phi, 0).transpose() * gradV;
                        double gradrho = pbl->fluid->rho->evalGrad(*e, isol->phi, 0) *
                                         e->computeGradient(isol->phi, 0).transpose() *
                                         gradV;
                        T_dM_x.push_back(Eigen::Triplet<double>(
                            jj, e->nodes[inodOnElm]->row * nDim + idim, -gradM / nAdjElms));
                        T_drho_x.push_back(Eigen::Triplet<double>(
                            jj, e->nodes[inodOnElm]->row * nDim + idim,
                            -gradrho / nAdjElms));
                        for (int idim2 = 0; idim2 < nDim; idim2++)
                            T_dv_x.push_back(Eigen::Triplet<double>(
                                jj * nDim + idim2, e->nodes[inodOnElm]->row * nDim + idim,
                                -gradV(idim2) / nAdjElms));
                    }
            }
            ++jj;
        }
    }
    dRM_x.setFromTriplets(T_dM_x.begin(), T_dM_x.end());
    dRrho_x.setFromTriplets(T_drho_x.begin(), T_drho_x.end());
    dRv_x.setFromTriplets(T_dv_x.begin(), T_dv_x.end());
    dRM_x.prune(0.);
    dRrho_x.prune(0.);
    dRv_x.prune(0.);
}

/**
 * @brief Run the viscous solver and extract the deltaStar
 * on all the stations
 *
 * @return Eigen::VectorXd deltaStar
 */
Eigen::VectorXd CoupledAdjoint::runViscous()
{
    int exitCode = vsol->run();
    if (exitCode != 0 && vsol->verbose > 0)
        std::cout << "vsol terminated with exit code " << exitCode << std::endl;

    // Extract deltaStar
    std::vector<Eigen::VectorXd> dStar_p;

    // Extract parts in a loop
    for (size_t i = 0; i < vsol->sections[0].size(); ++i)
    {
        std::vector<double> deltaStar_vec = this->vsol->sections[0][i]->deltaStar;
        Eigen::VectorXd deltaStar_eigen =
            Eigen::Map<Eigen::VectorXd>(deltaStar_vec.data(), deltaStar_vec.size());
        dStar_p.push_back(deltaStar_eigen);
    }
    // Calculate the total size
    int nNs = 0;
    for (const auto &p : dStar_p)
        nNs += p.size();

    // Concatenate the vectors
    Eigen::VectorXd deltaStar(nNs);
    int idx = 0;
    for (const auto &p : dStar_p)
    {
        deltaStar.segment(idx, p.size()) = p;
        idx += p.size();
    }
    return deltaStar;
}

/**
 * @brief Transpose all matrices of the adjoint problem included in the system.
 * Not the others.
 */
void CoupledAdjoint::transposeMatrices()
{

    dRphi_phi = dRphi_phi.transpose();
    dRphi_M = dRphi_M.transpose();
    dRphi_rho = dRphi_rho.transpose();
    dRphi_v = dRphi_v.transpose();
    dRphi_dStar = dRphi_dStar.transpose();
    dRphi_blw = dRphi_blw.transpose();

    dRM_phi = dRM_phi.transpose();
    dRM_M = dRM_M.transpose();
    dRM_rho = dRM_rho.transpose();
    dRM_v = dRM_v.transpose();
    dRM_dStar = dRM_dStar.transpose();
    dRM_blw = dRM_blw.transpose();

    dRrho_phi = dRrho_phi.transpose();
    dRrho_M = dRrho_M.transpose();
    dRrho_rho = dRrho_rho.transpose();
    dRrho_v = dRrho_v.transpose();
    dRrho_dStar = dRrho_dStar.transpose();
    dRrho_blw = dRrho_blw.transpose();

    dRv_phi = dRv_phi.transpose();
    dRv_M = dRv_M.transpose();
    dRv_rho = dRv_rho.transpose();
    dRv_v = dRv_v.transpose();
    dRv_dStar = dRv_dStar.transpose();
    dRv_blw = dRv_blw.transpose();

    dRdStar_phi = dRdStar_phi.transpose();
    dRdStar_M = dRdStar_M.transpose();
    dRdStar_rho = dRdStar_rho.transpose();
    dRdStar_v = dRdStar_v.transpose();
    dRdStar_dStar = dRdStar_dStar.transpose();
    dRdStar_blw = dRdStar_blw.transpose();

    dRblw_phi = dRblw_phi.transpose();
    dRblw_M = dRblw_M.transpose();
    dRblw_rho = dRblw_rho.transpose();
    dRblw_v = dRblw_v.transpose();
    dRblw_dStar = dRblw_dStar.transpose();
    dRblw_blw = dRblw_blw.transpose();
}

/**
 * @brief Write the matrices to files
 * @param matrix Eigen matrix to write
 * @param filename Name of the file to write
 * @return void
 *
 * The matrices are written in a file with the following format:
 * - The first line contains the column headers
 * - The following lines contain the matrix data
 * - The matrix data is written with a precision of 12 digits
 */
void CoupledAdjoint::writeMatrixToFile(const Eigen::MatrixXd &matrix,
                                       const std::string &filename)
{
    std::ofstream file(filename);
    if (file.is_open())
    {
        // Write the column headers
        for (int j = 0; j < matrix.cols(); ++j)
        {
            file << std::setw(15) << j;
            if (j != matrix.cols() - 1)
            {
                file << " ";
            }
        }
        file << "\n";

        // Write the matrix data
        for (int i = 0; i < matrix.rows(); ++i)
        {
            for (int j = 0; j < matrix.cols(); ++j)
            {
                file << std::fixed << std::setprecision(12) << std::setw(15)
                     << matrix(i, j);
                if (j != matrix.cols() - 1)
                {
                    file << " ";
                }
            }
            file << "\n";
        }
        file.close();
    }
}
